'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.mkModule = mkModule;
exports.mkRoute = mkRoute;
exports.mkComponent = mkComponent;

var _lodash = require('lodash');

var _lodash2 = _interopRequireDefault(_lodash);

var _fsExtra = require('fs-extra');

var _fsExtra2 = _interopRequireDefault(_fsExtra);

var _path = require('path');

var _path2 = _interopRequireDefault(_path);

var _async = require('async');

var _async2 = _interopRequireDefault(_async);

var _colors = require('colors');

var _colors2 = _interopRequireDefault(_colors);

var _utils = require('./utils.js');

var _variables = require('../conf/variables');

var VARS = _interopRequireWildcard(_variables);

var _option_merger = require('./option_merger');

var _option_merger2 = _interopRequireDefault(_option_merger);

function _interopRequireWildcard(obj) { if (obj && obj.__esModule) { return obj; } else { var newObj = {}; if (obj != null) { for (var key in obj) { if (Object.prototype.hasOwnProperty.call(obj, key)) newObj[key] = obj[key]; } } newObj.default = obj; return newObj; } }

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

/**
 * 生成模块
 * @param  {string}  name    模块名称
 * @param  {Object}  datas   数据
 * @param  {Boolean} isForce 是否强制
 * @return {Boolean}
 */
function mkModule(name, options, callback) {
  /* istanbul ignore if */
  if (3 > arguments.length) {
    return mkModule(name, {}, options);
  }

  /* istanbul ignore if */
  if (!_lodash2.default.isFunction(callback)) {
    throw new Error('Callback is not provided.');
  }

  /**
   * 检查模板是否存在, 不存在则报错并退出
   * check moudle template exists and exist process when template not exists.
   */
  var templateDir = _path2.default.join(_option_merger2.default.EXEC_PATH, './templates/module');
  if (!_fsExtra2.default.existsSync(templateDir)) {
    callback(new Error('Module template is not found, see ' + templateDir));
    return;
  }

  options = _lodash2.default.defaults(options, {
    ignoreExists: false,
    basePath: process.cwd(),
    distFolder: _path2.default.join(VARS.RESOURCE_FOLDER_NAME, VARS.ENTRY_FOLDER_NAME)
  });

  var names = (0, _utils.convertName)(name);
  var filename = names.underscore;
  var moduleDir = _path2.default.join(options.basePath, options.distFolder, filename);

  /**
   * 检查是否已经存在, 如果模块已经存在则直接退出
   * check module exists and exit process when file is exists.
   */
  if (_fsExtra2.default.existsSync(moduleDir)) {
    true !== options.ignoreExists && (0, _utils.trace)(('Module ' + _colors2.default.bold(name) + ' is already exists.').yellow);
    callback(null);
    return;
  }

  /**
   * 复制并渲染模板文件
   * copy and render file with data.
   */
  _fsExtra2.default.ensureDir(moduleDir, function (error) {
    if (error) {
      callback(error);
      return;
    }

    (0, _utils.copyAndRender)(templateDir, moduleDir, { names: names }, callback);
  });
}

/**
 * 创建模块与组件
 * @param  {String} argv 指令
 */
function mkRoute(route, moduleName, options, callback) {
  if (!_lodash2.default.isFunction(callback)) {
    throw new Error('Callback is not provided.');
  }

  var routes = _lodash2.default.isArray(route) ? route : route.split('\/');
  var family = [moduleName];
  var tasks = _lodash2.default.map(routes, function (name) {
    return function (callback) {
      mkComponent(name, family, options, function (error, stats) {
        if (error) {
          callback(error);
          return;
        }

        family.push(name);

        callback(null, stats);
      });
    };
  });

  _async2.default.series(tasks, function (error, stats) {
    if (error) {
      callback(error);
      return;
    }

    stats = _lodash2.default.flattenDeep(stats);
    stats = _lodash2.default.filter(stats);

    callback(null, stats);
  });
}

/**
 * 创建组件
 * @param  {String} name   组件名称
 * @param  {Array}  family 模块关系
 * @param  {Object} datas  数据
 * @return {Boolean}
 */
function mkComponent(name, family, options, callback) {
  /* istanbul ignore if */
  if (4 > arguments.length) {
    return mkComponent(name, family, {}, options);
  }

  /* istanbul ignore if */
  if (!_lodash2.default.isFunction(callback)) {
    throw new Error('Callback is not provided.');
  }

  /**
   * 检查模板是否存在, 不存在则报错并退出
   * check component template exists and exist process when template not exists.
   */
  var templateDir = _path2.default.join(_option_merger2.default.EXEC_PATH, './templates/component');
  if (!_fsExtra2.default.existsSync(templateDir)) {
    callback(new Error('Component template is not found, see ' + templateDir + '.'));
    return;
  }

  options = _lodash2.default.defaults(options, {
    ignoreExists: true,
    basePath: process.cwd(),
    distFolder: _path2.default.join(VARS.RESOURCE_FOLDER_NAME, VARS.ENTRY_FOLDER_NAME)
  });

  var names = (0, _utils.convertName)(name);
  var pwd = _lodash2.default.map(family, function (name) {
    return name + '/components/';
  });

  var dist = _path2.default.join(options.basePath, options.distFolder, pwd.join('\/'), name);
  if (_fsExtra2.default.existsSync(dist)) {
    true !== options.ignoreExists && (0, _utils.trace)(('Component ' + _colors2.default.bold(name) + ' is already exists.').yellow);
    callback(null);
    return;
  }

  /**
   * 复制并渲染模板文件
   * copy and render file with data.
   */
  _fsExtra2.default.ensureDir(dist, function () {
    var nsNames = _lodash2.default.map(family, _utils.convertName);
    var ns = {
      camelcase: _lodash2.default.map(nsNames, 'camelcase').join('.'),
      underscore: _lodash2.default.map(nsNames, 'underscore').join('/'),
      hyphen: _lodash2.default.map(nsNames, 'hyphen').join(' '),
      cssFamily: _lodash2.default.map(nsNames, function (_ref) {
        var hyphen = _ref.hyphen;

        return hyphen + '-viewport';
      })
    };

    (0, _utils.copyAndRender)(templateDir, dist, { names: names, ns: ns }, callback);
  });

  return;
}