'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.convertName = convertName;
exports.formatBytes = formatBytes;
exports.resolvePath = resolvePath;
exports.trace = trace;
exports.buildInheritance = buildInheritance;
exports.copyAndRender = copyAndRender;
exports.printStats = printStats;

var _lodash = require('lodash');

var _lodash2 = _interopRequireDefault(_lodash);

var _fsExtra = require('fs-extra');

var _fsExtra2 = _interopRequireDefault(_fsExtra);

var _path = require('path');

var _path2 = _interopRequireDefault(_path);

var _async = require('async');

var _async2 = _interopRequireDefault(_async);

var _colors = require('colors');

var _colors2 = _interopRequireDefault(_colors);

var _handlebars = require('handlebars');

var _handlebars2 = _interopRequireDefault(_handlebars);

var _columnify = require('columnify');

var _columnify2 = _interopRequireDefault(_columnify);

var _option_merger = require('./option_merger');

var _option_merger2 = _interopRequireDefault(_option_merger);

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

var ingoreTrace = -1 === _lodash2.default.indexOf(process.argv, '--quiet');

/**
 * convert name
 * @param  {String} name
 * @return {Object}
 * @description
 * convert "-", "_", "camelName", "NAME", "name"
 * "-hyphen" for style class
 * "_underscore" for filename
 * "camelName" for javascript variables
 */
function convertName(name) {
  var camelcase = name.replace(/[- _]([\w])/g, function ($all, $1) {
    return $1.toUpperCase();
  }).replace(/^[A-Z]/, function ($all) {
    return $all.toLowerCase();
  });

  var underscore = camelcase.replace(/[A-Z]/g, function ($all) {
    return '_' + $all.toLowerCase();
  });

  var hyphen = camelcase.replace(/[A-Z]/g, function ($all) {
    return '-' + $all.toLowerCase();
  });

  var blank = camelcase.replace(/[A-Z]/g, function ($all) {
    return ' ' + $all.toLowerCase();
  }).replace(/^[a-z]/, function ($all) {
    return $all.toUpperCase();
  });

  var upCamelcase = camelcase.replace(/^[a-z]/, function ($all) {
    return $all.toUpperCase();
  });

  return {
    camelcase: camelcase,
    upCamelcase: upCamelcase,
    underscore: underscore,
    hyphen: hyphen,
    blank: blank
  };
}

/**
 * format size by unit
 * @param  {Number} bytes    size
 * @param  {Number} decimals
 * @return {String}
 */
function formatBytes(bytes, decimals) {
  if (0 === bytes) {
    return '0 Bytes';
  }

  var k = 1000;
  var dm = decimals + 1 || 3;
  var sizes = ['Bytes', 'KB', 'MB', 'GB', 'TB', 'PB', 'EB', 'ZB', 'YB'];
  var i = Math.floor(Math.log(bytes) / Math.log(k));

  return parseFloat((bytes / Math.pow(k, i)).toFixed(dm)) + ' ' + sizes[i];
}

/**
 * check path is absolute or not.
 * if not, it will join cwd path in front of path.
 * @param  {String} file path
 * @param  {String} cwd  addition path
 * @return {String}
 */
function resolvePath() {
  var file = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : './';
  var cwd = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : process.cwd();

  if (_path2.default.isAbsolute(file)) {
    return file;
  }

  return _path2.default.join(cwd, file);
}

/**
 * print stats
 * @param  {Object} options setting
 */
function trace(message) {
  ingoreTrace && console.log(message);
}

/**
 * build a class with some prototypes use for inherit
 * @param  {Object} proto prototypes
 * @return {Class}
 */
function buildInheritance(proto) {
  var InhertClass = function InhertClass() {};
  Object.assign(InhertClass.prototype, proto);
  return InhertClass;
}

/**
 * 复制并渲染
 * @param {Array}  files   文件集合
 * @param {Object} datas   渲染的数据
 * @param {String} fromDir 资源所在路径
 * @param {String} toDir   目标路径
 */
function copyAndRender() {
  var fromDir = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : '';
  var toDir = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : '';
  var datas = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : {};
  var callback = arguments[3];

  /* istanbul ignore if */
  if (!_lodash2.default.isFunction(callback)) {
    throw new Error('Callback is not provided.');
  }

  var files = _fsExtra2.default.readdirSync(fromDir);

  if (_lodash2.default.isEmpty(files)) {
    callback(null, []);
    return;
  }

  var tasks = _lodash2.default.map(files, function (filename) {
    return function (callback) {
      var file = _path2.default.join(fromDir, filename);

      /**
       * 判断文件目录是否为文件夹, 如果为文件夹则创建文件夹并将
       * 里面的文件全部复制到目标相应目录
       * 这里使用地推方法, 当所有文件都复制完毕会退出递归
       */
      if (_fsExtra2.default.statSync(file).isDirectory()) {
        var targetDir = _path2.default.join(toDir, filename);

        _fsExtra2.default.ensureDir(targetDir, function (error) {
          if (error) {
            callback(error);
            return;
          }

          copyAndRender(file, targetDir, datas, callback);
        });

        return;
      }

      var targetFile = _path2.default.join(toDir, filename);

      /**
       * 模板文件不是 handlebars 文件则退出,
       * 目前暂只支持 handlebars 的模板引擎.
       */
      if ('.hbs' !== _path2.default.extname(file)) {
        /**
         * 如果目标文件已经存在, 则退出不做任何操作,
         * 因此请确定文件是否存在, 若存在则无办法继续执行复制.
         */
        if (_fsExtra2.default.existsSync(targetFile)) {
          callback(null);
          return;
        }

        _fsExtra2.default.copy(file, targetFile, function (error) {
          if (error) {
            callback(error);
            return;
          }

          callback(null, { assets: targetFile, size: _fsExtra2.default.statSync(targetFile).size });
        });

        return;
      }

      var doneFile = targetFile.replace(/\.hbs$/, '');

      /**
       * 如果目标文件已经存在, 则退出不做任何操作,
       * 因此请确定文件是否存在, 若存在则无办法继续执行复制.
       */
      if (_fsExtra2.default.existsSync(doneFile)) {
        callback(null);
        return;
      }

      /**
       * 编译文件并保存到相应的文件目录
       */
      var template = _fsExtra2.default.readFileSync(file, 'utf-8');
      var compile = _handlebars2.default.compile(template);
      var source = compile(datas);

      _fsExtra2.default.writeFileSync(doneFile, source);
      callback(null, { assets: doneFile, size: source.length });
    };
  });

  _async2.default.parallel(tasks, function (error, stats) {
    if (error) {
      callback(error);
      return;
    }

    stats = _lodash2.default.flattenDeep(stats);
    stats = _lodash2.default.filter(stats);

    callback(null, stats);
  });
}

/**
 * Print results
 * @param  {Array}  stats   result set
 * @param  {Object} options columnify setting
 */
function printStats(stats, options) {
  /* istanbul ignore if */
  if (_lodash2.default.isEmpty(stats)) {
    trace(_colors2.default.yellow('Generate completed but nothing to be generated.'));
    return false;
  }

  options = _lodash2.default.defaultsDeep(options, {
    headingTransform: function headingTransform(heading) {
      return (heading.charAt(0).toUpperCase() + heading.slice(1)).white.bold;
    },

    config: {
      assets: {
        align: 'right',
        dataTransform: function dataTransform(file) {
          file = file.replace(_option_merger2.default.ROOT_PATH + '/', '');
          return _colors2.default.green(file).bold;
        }
      },
      size: {
        align: 'right',
        dataTransform: function dataTransform(size) {
          return formatBytes(size);
        }
      },
      domain: {
        align: 'right',
        dataTransform: function dataTransform(domain) {
          domain = _lodash2.default.isArray(domain) ? domain.join(',') : domain;
          domain = _colors2.default.green(domain);
          return _colors2.default.bold(domain);
        }
      },
      proxy: {
        align: 'right'
      },
      entries: {
        align: 'left',
        dataTransform: function dataTransform(entries) {
          entries = _lodash2.default.isArray(entries) ? entries.join('|') : entries;
          entries = _colors2.default.green(entries);
          return _colors2.default.bold(entries);
        }
      }
    }
  });

  /* eslint no-console:off */
  trace((0, _columnify2.default)(stats, options) + '\n');
  return true;
}